/*****************************************************************
 * Copyright (c) 2015 Texas Instruments and others
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Raymond Pang - Initial API and implementation
 *****************************************************************/
var gc = gc || {};
gc.databind = gc.databind || {};


(function() // closure for private static methods and data.
{
	var MqttPahoTransport = function() 
	{
		this.connected = false;
	  	gc.databind.AbstractMqttTransport.call(this);
	};
	
	MqttPahoTransport.prototype = new gc.databind.AbstractMqttTransport();

	MqttPahoTransport.prototype.connect = function( conn )
	{
		var error = this.disconnect();
		
		if ( error !== '' )
			return error;
		
		var options = {
			timeout: 5,
			useSSL: true,
			userName: '',
			password: '',
			keepAliveInterval: 60,
			cleanSession: true,
		   	onSuccess: onConnect,
		   	onFailure: onFailure,
		};

		if ( conn.timeout !== undefined )  			{ options.timeout = conn.timeout; }
		if ( conn.useSSL !== undefined )  			{ options.useSSL = conn.useSSL; }
		if ( conn.userName !== undefined )  		{ options.userName = conn.userName; }
		if ( conn.password !== undefined )  		{ options.password = conn.password; }
		if ( conn.keepAliveInterval !== undefined )	{ options.keepAliveInterval = conn.keepAliveInterval; }
		if ( conn.cleanSession !== undefined )  	{ options.cleanSession = conn.cleanSession; }

		var port = parseInt( conn.port ) || 443;
		
  		this.client = new Paho.MQTT.Client( conn.host, port, "", conn.clientid );
	   	this.client.onMessageArrived = conn.onMessageArrived;
		this.client.onMessageDelivered = conn.onMessageDelivered;
   		this.client.onConnectionLost = onConnectionLost;
   		
   		try
   		{
   			this.client.connect(options);			   
   		}
   		catch( e )
   		{
   			error = e;
   		};		 
   		
   		var that = this;
   		function onConnect() 
   		{
   			that.connected = true;
   			if ( conn.onConnect )
   				conn.onConnect();
   		}	
   		
   		function onConnectionLost( responseObject ) 
   		{
   			that.connected = false;
   			if ( conn.onConnectionLost )
   			{
   	  			var error = '';
				if ( responseObject.errorCode != 0 )
					error += ": " + responseObject.errorMessage;
				conn.onConnectionLost( error );
    		}
   		}
   		
		function onFailure(responseObject) 
		{
			if ( conn.onFailure )
			{
				conn.onFailure( "Connection failure: " + responseObject.errorMessage );
			}
		}
		  
  		return error;
	};
	
	MqttPahoTransport.prototype.disconnect = function()
	{
		var error = '';
		if ( this.client && this.connected )
		{
			try
			{
				this.client.disconnect();
			}
			catch( e )
			{
				error =  e;
			};
		}
	   return error;
	};
	
	MqttPahoTransport.prototype.subscribe = function( topic, opt )
	{
		var error = '';
		if ( this.client && this.connected )
		{
			var options = {
 			   	onFailure: onFailure
   		 	};
			if ( opt.onSuccess )
			{
				options.onSuccess = opt.onSuccess;
			}
			options.timeout = opt.timeout || 5,
			options.qos = opt.qos || 0;
			try
			{
				this.client.subscribe( topic, options );
			}
			catch( e )
			{
				error =  e;
			};
		}
		function onFailure(responseObject) 
		{
			if ( opt.onFailure )
			{
				opt.onFailure( "Subscription failure:" + topic + ":" + responseObject.errorMessage );
			}
		}
		return error;
	};
	
	MqttPahoTransport.prototype.unsubscribe = function( topic, opt )
	{
		var error = '';
		if ( this.client && this.connected )
		{
			var options = {
	 			   	onFailure: onFailure
	   		 };
			if ( opt.onSuccess )
			{
				options.onSuccess = opt.onSuccess;
			}
			options.timeout = opt.timeout || 5;
			try
			{
				this.client.unsubscribe( topic, options );
			}
			catch( e )
			{
				error =  e;
			};
			function onFailure(responseObject) 
			{
				if ( opt.onFailure )
				{
					opt.onFailure( "Unsubscription failure:" + topic + ":" + responseObject.errorMessage );
				}
			}
		}
		return error;
	};
	
	MqttPahoTransport.prototype.publish = function( topic, payload, opt )
	{
		var error = '';
		if ( this.client && this.connected )
		{
		   	var message = new Paho.MQTT.Message(JSON.stringify(payload));
		   	message.destinationName = topic; 
		   	message.qos = opt.qos || 0;
		   	message.retained = opt.retained || false;
			try
			{
				this.client.send(message);
			}
			catch( e )
			{
				error =  e;
			};
			return error;
		}
	};
	
	
	//--- via Cloud
	
	var MqttCloudTransport = function() 
	{
		gc.databind.AbstractMqttTransport.call(this);
	};
	
	MqttCloudTransport.prototype = new gc.databind.AbstractMqttTransport();
	
	MqttCloudTransport.prototype.connect = function( conn )
	{
		var error = this.disconnect();
		
		this.connected = false;
		this.ws_connected = false;
		
		if ( error !== '' )
			return error;
		
		var wsURL = ( window.location.protocol.indexOf("https") === 0  ? "wss" : "ws" )
				+ "://" + window.location.hostname + "/mqtt";
		
		try
		{
			if ( this.ws )
			{
				this.ws.onmessage = null;
				this.ws.onerror = null;
				this.ws.onclose = null;
				this.ws = null;
			}
			this.ws = new WebSocket( wsURL, "ti-cloud-mqtt" );
		}
		catch(e)
		{
			return e;
		};
		
		var that = this;
		this.ws.onopen = function(event)
		{
			var opts = {
					host: conn.host,
					port:  conn.port,
					clientid: conn.clientid,
					timeout: 5,
					useSSL: conn.useSSL,
					userName: '',
					password: '',
					keepAliveInterval: 60,
					cleanSession: true,
			};
			if ( conn.timeout !== undefined )  			{ opts.timeout = conn.timeout; }
			if ( conn.userName !== undefined )  		{ opts.userName = conn.userName; }
			if ( conn.password !== undefined )  		{ opts.password = conn.password; }
			if ( conn.keepAliveInterval !== undefined )	{ opts.keepAliveInterval = conn.keepAliveInterval; }
			if ( conn.cleanSession !== undefined )  	{ opts.cleanSession = conn.cleanSession; }

			that.ws_connected = true;
			that._send( 'connect', opts );
		};
		
		this.ws.onmessage = function(msg)
		{
        	var event = JSON.parse( msg.data );
        	
        	if ( event.name === 'connected' )
        	{
        		that.connected = true;
       			if ( conn.onConnect )
       				conn.onConnect();
        	}
        	else if ( event.name === 'close' )
        	{
        		that.connected = false;
       			if ( conn.onConnectionLost )
       				conn.onConnectionLost();
        	}
        	else if ( event.name === 'error' )
        	{
       			if ( conn.onFailure )
       				conn.onFailure(event.detail.msg);
        	}
        	else if ( event.name === 'messageArrived' )
        	{
        			if ( conn.onMessageArrived )
       			{
       				var message = { 
       						destinationName: event.detail.dest,
       						payloadString: event.detail.payload
       				};
       				conn.onMessageArrived(message);
      			}
        	}
		};
		
		this.ws.onerror = function(event)
		{
   			if ( conn.onFailure )
   			{
				conn.onFailure( "WS Connection failure: " + event.reason );
			}
		};
		
		this.ws.onclose = function(event)
		{
			if ( that.ws_connected  )
			{
				that.ws_connected = false;
				that.connected = false;
	   			if ( conn.onConnectionLost && event.code != 1000 )
	   			{
	   				var error = "WS Connection Closed: clean:" + (event.wasClean ? 'Y' : 'N') 
							+ " code:" + event.code.toString() + " reason:" + event.reason;
					conn.onConnectionLost( error );
	    		}				
			}
		};
 		return error;
	};
	
	MqttCloudTransport.prototype._send = function( cmdName, data )
	{
		var error = '';
		try
		{
			var msg = {	cmd: cmdName };
			if ( data )
				msg.data = data;
			this.ws.send( JSON.stringify( msg ) );
		}
		catch( e )
		{
			error =  e;
		};
		return error;
	};
	
	MqttCloudTransport.prototype.disconnect = function()
	{
		var error = '';
		if ( this.ws && this.ws_connected )
		{
			if ( this.connected )
			{
				error = this._send( 'disconnect' );
			}
			else
			{
				this.ws.close();
			}
		}
	   return error;
	};
	
	MqttCloudTransport.prototype.subscribe = function( topic, opt )
	{
		var error = '';
		if ( this.ws && this.connected )
		{
			error = this._send( 'subscribe', { dest: topic, options: opt } );
		}
		return error;
	};
	
	MqttCloudTransport.prototype.unsubscribe = function( topic, opt )
	{
		var error = '';
		if ( this.ws && this.connected )
		{
			error = this._send( 'unsubscribe', { dest: topic, options: opt } );
		}
		return error;
	};
	
	MqttCloudTransport.prototype.publish = function( topic, payload, opt )
	{
		var error = '';
		if ( this.ws && this.connected )
		{
			error = this._send( 'publish', { dest: topic, msg: payload, options: opt } );
		}
		return error;
	};

	//--- Hybrid  wss -> paho    ws -> Cloud 
	
	var MqttHybridTransport = function() 
	{
		gc.databind.AbstractMqttTransport.call(this);
	};
	
	MqttHybridTransport.prototype = new gc.databind.AbstractMqttTransport();
	
	MqttHybridTransport.prototype.connect = function( conn )
	{
		var error = "Internal error: invalid mqtt transport";
		if ( this.transport )
		{
			this.transport.disconnect();
		}
		var usePaho = conn.useSSL;
		this.transport = gc.databind.CreateMqttTrasport( usePaho ? "paho" : "cloud" );
		if ( this.transport )
		{
			error = this.transport.connect( conn );
		}
		return error;
	};
	
	MqttHybridTransport.prototype.disconnect = function()
	{
		var error = "Internal error: invalid mqtt transport";
		if ( this.transport )
		{
			error = this.transport.disconnect();
		}
		return error;
	};
	
	MqttHybridTransport.prototype.subscribe = function( topic, opt )
	{
		var error = "Internal error: invalid mqtt transport";
		if ( this.transport )
		{
			error = this.transport.subscribe( topic, opt );
		}
		return error;
	};
	
	MqttHybridTransport.prototype.unsubscribe = function( topic, opt )
	{
		var error = "Internal error: invalid mqtt transport";
		if ( this.transport )
		{
			error = this.transport.unsubscribe( topic, opt );
		}
		return error;
	};
	
	MqttHybridTransport.prototype.publish = function( topic, payload, opt )
	{
		var error = "Internal error: invalid mqtt transport";
		if ( this.transport )
		{
			error = this.transport.publish( topic, payload, opt );
		}
		return error;
	};
	
	gc.databind.CreateMqttTrasport = function( name ) 
	{
		if ( name === "paho" )
			return new MqttPahoTransport();
		else if ( name === "cloud" )
			return new MqttCloudTransport();
		else if ( name === "hybrid" )
			return new MqttHybridTransport();
		return undefined;
	};
	
}());
